#include "StdAfx.h"
#include "Fat32Format.h"

static ULONG GetShiftCount(ULONG Value)
{
  ULONG i = 1;
  while (Value > 0)
    {
      i++;
      Value /= 2;
    }
  return i - 2;
}


static ULONG CalcVolumeSerialNumber(VOID)
{
  SYSTEMTIME SystemTime;
  ULONG Serial;
  PUCHAR Buffer;

  GetSystemTime(&SystemTime);

  Buffer = (PUCHAR)&Serial;
  Buffer[0] = (UCHAR)(SystemTime.wYear & 0xFF) + (UCHAR)(SystemTime.wHour & 0xFF);
  Buffer[1] = (UCHAR)(SystemTime.wYear >> 8) + (UCHAR)(SystemTime.wMinute & 0xFF);
  Buffer[2] = (UCHAR)(SystemTime.wMonth & 0xFF) + (UCHAR)(SystemTime.wSecond & 0xFF);
  Buffer[3] = (UCHAR)(SystemTime.wDay & 0xFF) + (UCHAR)(SystemTime.wMilliseconds & 0xFF);

  return Serial;
}

Fat32Format::Fat32Format(void)
{
	m_hPhysicalDisk = NULL;
}

Fat32Format::~Fat32Format(void)
{
	if(m_hPhysicalDisk)
	{
		CloseHandle(m_hPhysicalDisk);
		m_hPhysicalDisk = NULL;
	}
}

bool Fat32Format::Initialize(CString cstrDeviceName)
{
	DWORD errCode;
	DISK_GEOMETRY dg;
	ZeroMemory(&dg,sizeof(dg));
	DWORD cbReturned;

	if(m_hPhysicalDisk)
	{
		CloseHandle(m_hPhysicalDisk);
		m_hPhysicalDisk = NULL;
	}

	m_hPhysicalDisk=
		CreateFile(
		cstrDeviceName,
		GENERIC_READ|GENERIC_WRITE,
		FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
		NULL,
		OPEN_EXISTING,
		0,
		0
		);

	if(m_hPhysicalDisk==INVALID_HANDLE_VALUE) 
	{
		AfxMessageBox(CString(_T("Unable to open ")) + cstrDeviceName, MB_OK|MB_ICONEXCLAMATION);
		return 1;
	}
	else
	{
		errCode=
			DeviceIoControl(
			m_hPhysicalDisk,
			IOCTL_DISK_GET_DRIVE_GEOMETRY,
			NULL,0,
			&dg,sizeof(DISK_GEOMETRY),
			&cbReturned,
			NULL
			);

		if(!errCode) 
		{
			AfxMessageBox(CString(_T("Media Unplugged!!")), MB_OK|MB_ICONEXCLAMATION);
			return 1;
		} 
		else 
		{
			m_llVolumeSize = dg.Cylinders.QuadPart * dg.TracksPerCylinder * dg.SectorsPerTrack * dg.BytesPerSector;
		}
	}

	return 0;
}

int Fat32Format::SetPartition(int nIndex, DWORD size)
{
	DWORD errCode;
	DWORD cbReturned;

	// Get drive layout & partition information
	BYTE dliBuf[ sizeof(DRIVE_LAYOUT_INFORMATION)+
		(MAX_PARTITIONS-1)*sizeof(PARTITION_INFORMATION)
	];

	ZeroMemory(&dliBuf,sizeof(dliBuf));

	errCode=
		DeviceIoControl(
		m_hPhysicalDisk,
		IOCTL_DISK_GET_DRIVE_LAYOUT,
		NULL,0,
		&dliBuf,sizeof(dliBuf),
		&cbReturned,
		NULL
		);

	if(!errCode) {
		ErrorMessage(GetLastError());
		AfxMessageBox(_T("Error in IOCTL_DISK_GET_DRIVE_LAYOUT"), MB_OK|MB_ICONEXCLAMATION);
		return 1;
	} 
	else 
	{
		DRIVE_LAYOUT_INFORMATION& dli=
			*reinterpret_cast<DRIVE_LAYOUT_INFORMATION*>(dliBuf);

		int l = nIndex;
		dli.Signature = 0; //CalcVolumeSerialNumber();
		LONGLONG startOff = dli.PartitionEntry[l-1].StartingOffset.QuadPart + dli.PartitionEntry[l-1].PartitionLength.QuadPart;
		LONGLONG partLen = (LONGLONG)size*1024*1024;

		if(l == -1)
		{
			dli.PartitionEntry[2].HiddenSectors = dli.PartitionEntry[1].HiddenSectors;
			dli.PartitionEntry[2].PartitionType = dli.PartitionEntry[1].PartitionType;
			dli.PartitionEntry[2].PartitionLength.QuadPart = dli.PartitionEntry[1].PartitionLength.QuadPart;	
			dli.PartitionEntry[2].BootIndicator = dli.PartitionEntry[1].BootIndicator;
			dli.PartitionEntry[2].RecognizedPartition = dli.PartitionEntry[1].RecognizedPartition;
			dli.PartitionEntry[2].RewritePartition = TRUE;			
			dli.PartitionEntry[2].PartitionNumber = dli.PartitionEntry[1].PartitionNumber;
			dli.PartitionEntry[2].StartingOffset.QuadPart = dli.PartitionEntry[1].StartingOffset.QuadPart;

			dli.PartitionEntry[1].HiddenSectors = dli.PartitionEntry[0].HiddenSectors;
			dli.PartitionEntry[1].PartitionType = dli.PartitionEntry[0].PartitionType;
			dli.PartitionEntry[1].PartitionLength.QuadPart = dli.PartitionEntry[0].PartitionLength.QuadPart;	
			dli.PartitionEntry[1].BootIndicator = dli.PartitionEntry[0].BootIndicator;
			dli.PartitionEntry[1].RecognizedPartition = dli.PartitionEntry[0].RecognizedPartition;
			dli.PartitionEntry[1].RewritePartition = TRUE;			
			dli.PartitionEntry[1].PartitionNumber = dli.PartitionEntry[0].PartitionNumber;
			dli.PartitionEntry[1].StartingOffset.QuadPart = dli.PartitionEntry[0].StartingOffset.QuadPart;

			dli.PartitionEntry[0].BootIndicator = FALSE ;
			dli.PartitionEntry[0].RecognizedPartition = TRUE ;
			dli.PartitionEntry[0].RewritePartition = TRUE ;
			dli.PartitionEntry[0].HiddenSectors = 0;
			dli.PartitionEntry[0].PartitionType = PARTITION_FAT32_XINT13;
			//dli.PartitionEntry[0].PartitionNumber = 1;
			dli.PartitionEntry[0].StartingOffset.QuadPart = 1024*1024;
			dli.PartitionEntry[0].PartitionLength.QuadPart = partLen;
		}
		else if(l == -2)
		{
			dli.PartitionEntry[2].HiddenSectors = dli.PartitionEntry[1].HiddenSectors;
			dli.PartitionEntry[2].PartitionType = dli.PartitionEntry[1].PartitionType;
			dli.PartitionEntry[2].PartitionLength.QuadPart = dli.PartitionEntry[1].PartitionLength.QuadPart;	
			dli.PartitionEntry[2].BootIndicator = dli.PartitionEntry[1].BootIndicator;
			dli.PartitionEntry[2].RecognizedPartition = dli.PartitionEntry[1].RecognizedPartition;
			dli.PartitionEntry[2].RewritePartition = TRUE;			
			dli.PartitionEntry[2].PartitionNumber = dli.PartitionEntry[1].PartitionNumber;
			dli.PartitionEntry[2].StartingOffset.QuadPart = dli.PartitionEntry[1].StartingOffset.QuadPart;

			dli.PartitionEntry[1].HiddenSectors = 0;
			dli.PartitionEntry[1].PartitionType = PARTITION_FAT32_XINT13;
			dli.PartitionEntry[1].PartitionLength.QuadPart = partLen;	
			dli.PartitionEntry[1].BootIndicator = FALSE;
			dli.PartitionEntry[1].RecognizedPartition = TRUE;
			dli.PartitionEntry[1].RewritePartition = TRUE;			
			//dli.PartitionEntry[1].PartitionNumber = 3;
			dli.PartitionEntry[1].StartingOffset.QuadPart = dli.PartitionEntry[0].StartingOffset.QuadPart + dli.PartitionEntry[0].PartitionLength.QuadPart;
		}
		else if(l == 0)
		{
			dli.PartitionEntry[l].BootIndicator = FALSE ;
			dli.PartitionEntry[l].RecognizedPartition = TRUE ;
			dli.PartitionEntry[l].RewritePartition = TRUE ;
			dli.PartitionEntry[l].HiddenSectors = 0;
			dli.PartitionEntry[l].PartitionType = PARTITION_FAT32;
			dli.PartitionEntry[l].PartitionNumber = 1;
			dli.PartitionEntry[l].StartingOffset.QuadPart = 1024 * 1024;
			dli.PartitionEntry[l].PartitionLength.QuadPart = partLen;
		}
		else if(l == 3)
		{
			dli.PartitionCount = 8;

			dli.PartitionEntry[5].BootIndicator = FALSE ;
			dli.PartitionEntry[5].HiddenSectors = 0;
			dli.PartitionEntry[5].PartitionLength.QuadPart = 0;
			dli.PartitionEntry[5].PartitionNumber = 0;
			dli.PartitionEntry[5].PartitionType = PARTITION_ENTRY_UNUSED;
			dli.PartitionEntry[5].RecognizedPartition = FALSE ;
			dli.PartitionEntry[5].RewritePartition = TRUE ;
			dli.PartitionEntry[5].StartingOffset.QuadPart = 0;

			dli.PartitionEntry[6].BootIndicator = FALSE ;
			dli.PartitionEntry[6].HiddenSectors = 0;
			dli.PartitionEntry[6].PartitionLength.QuadPart = 0;
			dli.PartitionEntry[6].PartitionNumber = 0;
			dli.PartitionEntry[6].PartitionType = PARTITION_ENTRY_UNUSED;
			dli.PartitionEntry[6].RecognizedPartition = FALSE ;
			dli.PartitionEntry[6].RewritePartition = TRUE ;
			dli.PartitionEntry[6].StartingOffset.QuadPart = 0;

			dli.PartitionEntry[7].BootIndicator = FALSE ;
			dli.PartitionEntry[7].HiddenSectors = 0;
			dli.PartitionEntry[7].PartitionLength.QuadPart = 0;
			dli.PartitionEntry[7].PartitionNumber = 0;
			dli.PartitionEntry[7].PartitionType = PARTITION_ENTRY_UNUSED;
			dli.PartitionEntry[7].RecognizedPartition = FALSE ;
			dli.PartitionEntry[7].RewritePartition = TRUE ;
			dli.PartitionEntry[7].StartingOffset.QuadPart = 0;

			dli.PartitionEntry[4].BootIndicator = FALSE ;
			dli.PartitionEntry[4].RecognizedPartition = TRUE ;
			dli.PartitionEntry[4].RewritePartition = TRUE ;
			dli.PartitionEntry[4].HiddenSectors = 0;
			dli.PartitionEntry[4].PartitionType = PARTITION_FAT32;
			dli.PartitionEntry[4].PartitionNumber = dli.PartitionEntry[l-1].PartitionNumber + 1;
			dli.PartitionEntry[4].StartingOffset.QuadPart = startOff + 1024*1024;
			dli.PartitionEntry[4].PartitionLength.QuadPart = partLen;

			dli.PartitionEntry[l].BootIndicator = FALSE ;
			dli.PartitionEntry[l].RecognizedPartition = FALSE ;
			dli.PartitionEntry[l].RewritePartition = TRUE ;			
			dli.PartitionEntry[l].HiddenSectors = 0;
			dli.PartitionEntry[l].PartitionType = PARTITION_XINT13_EXTENDED;
			dli.PartitionEntry[l].PartitionNumber = 0;
			dli.PartitionEntry[l].StartingOffset.QuadPart = startOff;
			dli.PartitionEntry[l].PartitionLength.QuadPart = m_llVolumeSize - startOff;

		}	
		else if(l % 4 == 1 && l != 1)
		{
			dli.PartitionCount = l + 7;

			dli.PartitionEntry[l+4].BootIndicator = FALSE ;
			dli.PartitionEntry[l+4].HiddenSectors = 0;
			dli.PartitionEntry[l+4].PartitionLength.QuadPart = 0;
			dli.PartitionEntry[l+4].PartitionNumber = 0;
			dli.PartitionEntry[l+4].PartitionType = PARTITION_ENTRY_UNUSED;
			dli.PartitionEntry[l+4].RecognizedPartition = FALSE ;
			dli.PartitionEntry[l+4].RewritePartition = TRUE ;
			dli.PartitionEntry[l+4].StartingOffset.QuadPart = 0;

			dli.PartitionEntry[l+5].BootIndicator = FALSE ;
			dli.PartitionEntry[l+5].HiddenSectors = 0;
			dli.PartitionEntry[l+5].PartitionLength.QuadPart = 0;
			dli.PartitionEntry[l+5].PartitionNumber = 0;
			dli.PartitionEntry[l+5].PartitionType = PARTITION_ENTRY_UNUSED;
			dli.PartitionEntry[l+5].RecognizedPartition = FALSE ;
			dli.PartitionEntry[l+5].RewritePartition = TRUE ;
			dli.PartitionEntry[l+5].StartingOffset.QuadPart = 0;

			dli.PartitionEntry[l+6].BootIndicator = FALSE ;
			dli.PartitionEntry[l+6].HiddenSectors = 0;
			dli.PartitionEntry[l+6].PartitionLength.QuadPart = 0;
			dli.PartitionEntry[l+6].PartitionNumber = 0;
			dli.PartitionEntry[l+6].PartitionType = PARTITION_ENTRY_UNUSED;
			dli.PartitionEntry[l+6].RecognizedPartition = FALSE ;
			dli.PartitionEntry[l+6].RewritePartition = TRUE ;
			dli.PartitionEntry[l+6].StartingOffset.QuadPart = 0;

			dli.PartitionEntry[l + 3].BootIndicator = FALSE ;
			dli.PartitionEntry[l + 3].RecognizedPartition = TRUE ;
			dli.PartitionEntry[l + 3].RewritePartition = TRUE ;
			dli.PartitionEntry[l + 3].HiddenSectors = 0;
			dli.PartitionEntry[l + 3].PartitionType = PARTITION_FAT32;
			dli.PartitionEntry[l + 3].PartitionNumber = dli.PartitionEntry[l-1].PartitionNumber + 1;
			dli.PartitionEntry[l + 3].StartingOffset.QuadPart = startOff + 1024*1024;
			dli.PartitionEntry[l + 3].PartitionLength.QuadPart = partLen;

			dli.PartitionEntry[l].BootIndicator = FALSE ;
			dli.PartitionEntry[l].RecognizedPartition = FALSE ;
			dli.PartitionEntry[l].RewritePartition = TRUE ;			
			dli.PartitionEntry[l].HiddenSectors = 0;
			dli.PartitionEntry[l].PartitionType = PARTITION_EXTENDED;
			dli.PartitionEntry[l].PartitionNumber = 0;
			dli.PartitionEntry[l].StartingOffset.QuadPart = startOff;
			dli.PartitionEntry[l].PartitionLength.QuadPart = partLen + 1024*1024;
		}
		else
		{
			dli.PartitionEntry[l].BootIndicator = FALSE ;
			dli.PartitionEntry[l].RecognizedPartition = TRUE ;
			dli.PartitionEntry[l].RewritePartition = TRUE ;
			dli.PartitionEntry[l].HiddenSectors = 0;
			//dli.PartitionEntry[l].PartitionType = ((l==2) ? PARTITION_FAT32_XINT13 : PARTITION_FAT32);
			dli.PartitionEntry[l].PartitionType = PARTITION_FAT32_XINT13;
			dli.PartitionEntry[l].PartitionNumber = dli.PartitionEntry[l-1].PartitionNumber + 1;
			dli.PartitionEntry[l].StartingOffset.QuadPart = startOff;
			dli.PartitionEntry[l].PartitionLength.QuadPart = partLen;
		}

		errCode = DeviceIoControl( m_hPhysicalDisk, IOCTL_DISK_SET_DRIVE_LAYOUT,
			&dliBuf,sizeof(dliBuf), NULL, 0, &cbReturned,NULL );

		if(!errCode) {
			ErrorMessage(GetLastError());
			AfxMessageBox(_T("Error in IOCTL_DISK_SET_DRIVE_LAYOUT"), MB_OK|MB_ICONEXCLAMATION);
			return 1;
		}

	}

	return 0;
}

int Fat32Format::InitDrive()
{
	DWORD errCode;
	DWORD cbReturned;

	BYTE dliBuf[ sizeof(DRIVE_LAYOUT_INFORMATION)+
		(MAX_PARTITIONS-1)*sizeof(PARTITION_INFORMATION)];
	ZeroMemory(&dliBuf,sizeof(dliBuf));

	DRIVE_LAYOUT_INFORMATION& dli=
		*reinterpret_cast<DRIVE_LAYOUT_INFORMATION*>(dliBuf);

	dli.PartitionCount = 4;
	dli.PartitionEntry[0].RewritePartition = TRUE ;
	dli.PartitionEntry[1].RewritePartition = TRUE ;
	dli.PartitionEntry[2].RewritePartition = TRUE ;
	dli.PartitionEntry[3].RewritePartition = TRUE ;

	errCode = DeviceIoControl( m_hPhysicalDisk, IOCTL_DISK_SET_DRIVE_LAYOUT,
		&dliBuf,sizeof(dliBuf), NULL, 0, &cbReturned,NULL );

	if(!errCode) {
		ErrorMessage(GetLastError());
		AfxMessageBox(_T("Error in IOCTL_DISK_SET_DRIVE_LAYOUT"), MB_OK|MB_ICONEXCLAMATION);
		return 1;
	} 

	return 0;
}

int Fat32Format::DeletePartition(CString cstrDrive, int nIndex)
{
	//////////////////// p[eBṼu[gZN^NA
	//HANDLE fileHandle = CreateFile(cstrDrive, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE, 0, OPEN_EXISTING, 0, NULL);
	//if(fileHandle == INVALID_HANDLE_VALUE)
	//{
	//	ErrorMessage(GetLastError());
	//	AfxMessageBox(_T("Error in CreateFile ") + cstrDrive, MB_OK|MB_ICONEXCLAMATION);
	//	return 1;
	//}
	//DWORD dwBytesReturned;
	//if(!DeviceIoControl(
	//	fileHandle,              // handle to device
	//	FSCTL_LOCK_VOLUME,  // dwIoControlCode
	//	NULL,                          // lpInBuffer
	//	0,                             // nInBufferSize
	//	NULL,                          // lpOutBuffer
	//	0,                             // nOutBufferSize
	//	&dwBytesReturned,     // number of bytes returned
	//	NULL   // OVERLAPPED structure
	//	))
	//{
	//	ErrorMessage(GetLastError());
	//	AfxMessageBox(_T("FSCTL_LOCK_VOLUME failed!"), MB_OK|MB_ICONEXCLAMATION);
	//	return 1;
	//}

	//int nStatus = ERROR_SUCCESS;
	//FAT32_BOOT_SECTOR BootSector;
	//memset(&BootSector, 0, sizeof(FAT32_BOOT_SECTOR));
	//nStatus = Fat32WriteBootSector(fileHandle, &BootSector);
	//if (!NT_SUCCESS(nStatus))
	//{
	//	AfxMessageBox(_T("Fat32WriteBootSector() failed with status ") + nStatus, MB_OK|MB_ICONEXCLAMATION);
	//	return 1;
	//}
	//	DeviceIoControl(
	//	fileHandle,              // handle to device
	//	FSCTL_UNLOCK_VOLUME,  // dwIoControlCode
	//	NULL,                          // lpInBuffer
	//	0,                             // nInBufferSize
	//	NULL,                          // lpOutBuffer
	//	0,                             // nOutBufferSize
	//	&dwBytesReturned,     // number of bytes returned
	//	NULL   // OVERLAPPED structure
	//	);

	//CloseHandle(fileHandle);


	//////////////////// p[eBVe[uNA
	DWORD errCode;
	DWORD cbReturned;

	// Get drive layout & partition information
	BYTE dliBuf[ sizeof(DRIVE_LAYOUT_INFORMATION)+
		(MAX_PARTITIONS-1)*sizeof(PARTITION_INFORMATION)
	];

	ZeroMemory(&dliBuf,sizeof(dliBuf));

	errCode=
		DeviceIoControl(
		m_hPhysicalDisk,
		IOCTL_DISK_GET_DRIVE_LAYOUT,
		NULL,0,
		&dliBuf,sizeof(dliBuf),
		&cbReturned,
		NULL
		);

	if(!errCode) {
		ErrorMessage(GetLastError());
		AfxMessageBox(_T("Error in IOCTL_DISK_GET_DRIVE_LAYOUT"), MB_OK|MB_ICONEXCLAMATION);
		return 1;
	} 
	else 
	{
		DRIVE_LAYOUT_INFORMATION& dli=
			*reinterpret_cast<DRIVE_LAYOUT_INFORMATION*>(dliBuf);

		int l = nIndex;
		dli.PartitionEntry[l].BootIndicator = FALSE ;
		dli.PartitionEntry[l].HiddenSectors = 0;
		dli.PartitionEntry[l].PartitionLength.QuadPart = 0;
		dli.PartitionEntry[l].PartitionNumber = 0;
		dli.PartitionEntry[l].PartitionType = PARTITION_ENTRY_UNUSED;
		dli.PartitionEntry[l].RecognizedPartition = FALSE ;
		dli.PartitionEntry[l].RewritePartition = TRUE ;
		dli.PartitionEntry[l].StartingOffset.QuadPart = 0;

		if(l % 4 == 0 && l != 0)
		{
			dli.PartitionCount = l;
			if(l == 4)
			{
				dli.PartitionEntry[l-1].BootIndicator = FALSE ;
				dli.PartitionEntry[l-1].HiddenSectors = 0;
				dli.PartitionEntry[l-1].PartitionLength.QuadPart = 0;
				dli.PartitionEntry[l-1].PartitionNumber = 0;
				dli.PartitionEntry[l-1].PartitionType = PARTITION_ENTRY_UNUSED;
				dli.PartitionEntry[l-1].RecognizedPartition = FALSE ;
				dli.PartitionEntry[l-1].RewritePartition = TRUE ;
				dli.PartitionEntry[l-1].StartingOffset.QuadPart = 0;
			}
			else
			{
				dli.PartitionEntry[l-3].BootIndicator = FALSE ;
				dli.PartitionEntry[l-3].HiddenSectors = 0;
				dli.PartitionEntry[l-3].PartitionLength.QuadPart = 0;
				dli.PartitionEntry[l-3].PartitionNumber = 0;
				dli.PartitionEntry[l-3].PartitionType = PARTITION_ENTRY_UNUSED;
				dli.PartitionEntry[l-3].RecognizedPartition = FALSE ;
				dli.PartitionEntry[l-3].RewritePartition = TRUE ;
				dli.PartitionEntry[l-3].StartingOffset.QuadPart = 0;
			}
		}

		errCode = DeviceIoControl( m_hPhysicalDisk, IOCTL_DISK_SET_DRIVE_LAYOUT,
			&dliBuf,sizeof(dliBuf), NULL, 0, &cbReturned,NULL );

		if(!errCode) {
			ErrorMessage(GetLastError());
			AfxMessageBox(_T("Error in IOCTL_DISK_SET_DRIVE_LAYOUT"), MB_OK|MB_ICONEXCLAMATION);
			return 1;
		} 
	}

	return 0;
}

int Fat32Format::FormatPartition(CString cstrDrive, CString cstrVolumeLabel,int nIndex, BOOL bQuickFormat, CProgressCtrl *pProgFormat, HWND hParentDlg)
{
	DWORD errCode;
	DISK_GEOMETRY dg;
	ZeroMemory(&dg,sizeof(dg));
	DWORD cbReturned;
	m_hParentDlg = hParentDlg;

	errCode=
		DeviceIoControl(
		m_hPhysicalDisk,
		IOCTL_DISK_GET_DRIVE_GEOMETRY,
		NULL,0,
		&dg,sizeof(DISK_GEOMETRY),
		&cbReturned,
		NULL
		);

	if(!errCode) 
	{
		AfxMessageBox(_T("Error in IOCTL_DISK_GET_DRIVE_GEOMETRY"), MB_OK|MB_ICONEXCLAMATION);
		return 1;
	} 
	else 
	{
		// Get drive layout & partition information
		BYTE dliBuf[ sizeof(DRIVE_LAYOUT_INFORMATION)+
			(MAX_PARTITIONS-1)*sizeof(PARTITION_INFORMATION)
		];

		ZeroMemory(&dliBuf,sizeof(dliBuf));

		errCode=
			DeviceIoControl(
			m_hPhysicalDisk,
			IOCTL_DISK_GET_DRIVE_LAYOUT,
			NULL,0,
			&dliBuf,sizeof(dliBuf),
			&cbReturned,
			NULL
			);

		if(!errCode) {
			ErrorMessage(GetLastError());
			AfxMessageBox(_T("Error in IOCTL_DISK_GET_DRIVE_LAYOUT"), MB_OK|MB_ICONEXCLAMATION);
			return 1;
		} 
		else 
		{
			CloseHandle(m_hPhysicalDisk);
			m_hPhysicalDisk = NULL;

			DRIVE_LAYOUT_INFORMATION& dli=
				*reinterpret_cast<DRIVE_LAYOUT_INFORMATION*>(dliBuf);

			HANDLE fileHandle = CreateFile(cstrDrive, GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE, 0, OPEN_EXISTING, 0, NULL);
			if(fileHandle == INVALID_HANDLE_VALUE)
			{
				ErrorMessage(GetLastError());
				AfxMessageBox(_T("Error in CreateFile ") + cstrDrive, MB_OK|MB_ICONEXCLAMATION);
				return 1;
			}
			
			if(StartFormat(fileHandle, cstrVolumeLabel, &dli.PartitionEntry[nIndex], &dg, bQuickFormat, 0, pProgFormat))
			{
				AfxMessageBox(_T("Error in StartFormat"), MB_OK|MB_ICONEXCLAMATION);
				CloseHandle(fileHandle);
				return 1;
			}

			CloseHandle(fileHandle);
		}
	}
	return 0;
}


int Fat32Format::Fat32WriteBootSector(HANDLE FileHandle, PFAT32_BOOT_SECTOR BootSector)
{
	BOOL bRet;
	DWORD dwRet,dwError,dwBytes =0;
	PUCHAR NewBootSector;
	LARGE_INTEGER FileOffset;

	/* Allocate buffer for new bootsector */
	NewBootSector = (PUCHAR)LocalAlloc(LPTR, SECTORSIZE);
	if (NewBootSector == NULL)
	{
		ErrorMessage(GetLastError());
		AfxMessageBox(_T("Fat32WriteBootSector#LocalAlloc Failed"), MB_OK|MB_ICONEXCLAMATION);
		return 1;
	}

	/* Copy FAT32 BPB to new bootsector */
	//memcpy((NewBootSector + 3),
	//	&BootSector->OEMName[0],
	//	87); /* FAT32 BPB length (up to (not including) Res2) */

	memcpy(NewBootSector,
		&BootSector->magic0,
		SECTORSIZE); /* FAT32 BPB length (up to (not including) Res2) */

	/* Write sector 0 */
	FileOffset.QuadPart = 0ULL;
	dwRet = SetFilePointer(FileHandle ,FileOffset.LowPart, &FileOffset.HighPart, FILE_BEGIN);
	if (dwRet == INVALID_SET_FILE_POINTER && (dwError = GetLastError()) != NO_ERROR ) //changed!!
	{
		ErrorMessage(dwError);
		AfxMessageBox(_T("Fat32WriteBootSector#SetFilePointer1 Failed"), MB_OK|MB_ICONEXCLAMATION);
		return 1;
	}
	bRet = WriteFile(FileHandle, NewBootSector, SECTORSIZE, &dwBytes,NULL);

	if(!bRet || dwBytes != SECTORSIZE)
	{
		ErrorMessage(GetLastError());
		AfxMessageBox(_T("Fat32WriteBootSector#WriteFile1 Failed"), MB_OK|MB_ICONEXCLAMATION);
		LocalFree(NewBootSector);
		return 1;
	}

	/* Write backup boot sector */
	if (BootSector->BootBackup != 0x0000)
	{
		FileOffset.QuadPart = (ULONGLONG)((ULONG) BootSector->BootBackup * SECTORSIZE);
		DWORD dwRet = SetFilePointer(FileHandle ,FileOffset.LowPart, &FileOffset.HighPart, FILE_BEGIN);
		if (dwRet == INVALID_SET_FILE_POINTER && (dwError = GetLastError()) != NO_ERROR ) //changed!!
		{ 
			ErrorMessage(dwError);
			AfxMessageBox(_T("Fat32WriteBootSector#SetFilePointer2 Failed"), MB_OK|MB_ICONEXCLAMATION);
			return 1;
		}
		BOOL bRet = WriteFile(FileHandle, NewBootSector, SECTORSIZE, &dwBytes,NULL);

		if(!bRet || dwBytes != SECTORSIZE)
		{
			ErrorMessage(GetLastError());
			AfxMessageBox(_T("Fat32WriteBootSector#WriteFile2 Failed"), MB_OK|MB_ICONEXCLAMATION);
			LocalFree(NewBootSector);
			return 1;
		}
	}

	/* Free the new boot sector */
	LocalFree(NewBootSector);

	return ERROR_SUCCESS;
}


int Fat32Format::Fat32WriteFsInfo(HANDLE FileHandle, PFAT32_BOOT_SECTOR BootSector)
{
	BOOL bRet;
	DWORD dwRet,dwError,dwBytes =0;
	PFAT32_FSINFO FsInfo;
	LARGE_INTEGER FileOffset;

	/* Allocate buffer for new sector */
	FsInfo = (PFAT32_FSINFO)LocalAlloc(LPTR, BootSector->BytesPerSector);
	if (FsInfo == NULL)
	{
		ErrorMessage(GetLastError());
		AfxMessageBox(_T("Fat32WriteFsInfo#LocalAlloc Failed"), MB_OK|MB_ICONEXCLAMATION);
		return 1;
	}

	FsInfo->LeadSig = 0x41615252;
	FsInfo->StrucSig = 0x61417272;
	FsInfo->FreeCount = 0xffffffff;
	FsInfo->NextFree = 0xffffffff;
	FsInfo->TrailSig = 0xaa550000;

	/* Write sector */
	FileOffset.QuadPart = BootSector->FSInfoSector * BootSector->BytesPerSector;

	dwRet = SetFilePointer(FileHandle ,FileOffset.LowPart, &FileOffset.HighPart, FILE_BEGIN);
	if (dwRet == INVALID_SET_FILE_POINTER && (dwError = GetLastError()) != NO_ERROR ) //changed!!
	{
		ErrorMessage(dwError);
		AfxMessageBox(_T("Fat32WriteFsInfo#SetFilePointer Failed"), MB_OK|MB_ICONEXCLAMATION);
		return 1;
	}
	bRet = WriteFile(FileHandle, FsInfo, BootSector->BytesPerSector, &dwBytes,NULL);
	if(!bRet || dwBytes != BootSector->BytesPerSector)
	{
		ErrorMessage(GetLastError());
		AfxMessageBox(_T("Fat32WriteFsInfo#WriteFile Failed"), MB_OK|MB_ICONEXCLAMATION);
		LocalFree(FsInfo);
		return 1;
	}

	/* Free the new sector buffer */
	LocalFree(FsInfo);

	return ERROR_SUCCESS;
}


int Fat32Format::Fat32WriteFAT(HANDLE FileHandle, ULONG SectorOffset, PFAT32_BOOT_SECTOR BootSector)
{
	BOOL bRet;
	DWORD dwRet,dwError,dwBytes =0;
	PUCHAR Buffer;
	LARGE_INTEGER FileOffset;
	ULONG i;
	ULONG Size;
	ULONG Sectors;

	/* Allocate buffer */
	Buffer = (PUCHAR)LocalAlloc(LPTR, 64 * 1024);
	if (Buffer == NULL)
	{
		ErrorMessage(GetLastError());
		AfxMessageBox(_T("Fat32WriteFAT#LocalAlloc Failed"), MB_OK|MB_ICONEXCLAMATION);
		return 1;
	}

	/* FAT cluster 0 */
	Buffer[0] = 0xf8; /* Media type */
	Buffer[1] = 0xff;
	Buffer[2] = 0xff;
	Buffer[3] = 0x0f;
	/* FAT cluster 1 */
	Buffer[4] = 0xff; /* Clean shutdown, no disk read/write errors, end-of-cluster (EOC) mark */
	Buffer[5] = 0xff;
	Buffer[6] = 0xff;
	Buffer[7] = 0x0f;
	/* FAT cluster 2 */
	Buffer[8] = 0xff; /* End of root directory */
	Buffer[9] = 0xff;
	Buffer[10] = 0xff;
	Buffer[11] = 0x0f;

	/* Write first sector of the FAT */
	FileOffset.QuadPart = (SectorOffset + BootSector->ReservedSectors) * BootSector->BytesPerSector;

	dwRet = SetFilePointer(FileHandle ,FileOffset.LowPart, &FileOffset.HighPart, FILE_BEGIN);
	if (dwRet == INVALID_SET_FILE_POINTER && (dwError = GetLastError()) != NO_ERROR ) //changed!!
	{
		ErrorMessage(dwError);
		AfxMessageBox(_T("Fat32WriteFAT#SetFilePointer1 Failed"), MB_OK|MB_ICONEXCLAMATION);
		return 1;
	}
	bRet = WriteFile(FileHandle, Buffer, BootSector->BytesPerSector, &dwBytes,NULL);
	if(!bRet || dwBytes != BootSector->BytesPerSector)
	{
		ErrorMessage(GetLastError());
		AfxMessageBox(_T("Fat32WriteFAT#WriteFile1 Failed"), MB_OK|MB_ICONEXCLAMATION);
		LocalFree(Buffer);
		return 1;
	}

	/* Zero the begin of the buffer */
	memset(Buffer, 0, 12);

	/* Zero the rest of the FAT */
	Sectors = 64 * 1024 / BootSector->BytesPerSector;
	for (i = 1; i < BootSector->FATSectors32; i += Sectors)
	{
		/* Zero some sectors of the FAT */
		FileOffset.QuadPart = (SectorOffset + BootSector->ReservedSectors + i) * BootSector->BytesPerSector;
		Size = BootSector->FATSectors32 - i;
		if (Size > Sectors)
		{
			Size = Sectors;
		}
		Size *= BootSector->BytesPerSector;

		dwRet = SetFilePointer(FileHandle ,FileOffset.LowPart, &FileOffset.HighPart, FILE_BEGIN);
		if (dwRet == INVALID_SET_FILE_POINTER && (dwError = GetLastError()) != NO_ERROR ) //changed!!
		{
			ErrorMessage(dwError);
			AfxMessageBox(_T("Fat32WriteFAT#SetFilePointer2 Failed"), MB_OK|MB_ICONEXCLAMATION);
			return 1;
		}
		bRet = WriteFile(FileHandle, Buffer, Size, &dwBytes,NULL);
		if(!bRet /* || dwBytes != Size */)
		{
			ErrorMessage(GetLastError());
			AfxMessageBox(_T("Fat32WriteFAT#WriteFile2 Failed"), MB_OK|MB_ICONEXCLAMATION);
			LocalFree(Buffer);
			return 1;
		}
	}

	/* Free the buffer */
	LocalFree(Buffer);

	return ERROR_SUCCESS;
}


int Fat32Format::Fat32WriteRootDirectory(HANDLE FileHandle, PFAT32_BOOT_SECTOR BootSector)
{
	BOOL bRet;
	DWORD dwRet,dwError,dwBytes =0;
	PUCHAR Buffer;
	LARGE_INTEGER FileOffset;
	ULONGLONG FirstDataSector;
	ULONGLONG FirstRootDirSector;
	DirEntry aDirEntry;

	/* Allocate buffer for the cluster */
	Buffer = (PUCHAR)LocalAlloc(LPTR, BootSector->SectorsPerCluster * BootSector->BytesPerSector);
	if (Buffer == NULL)
	{
		ErrorMessage(GetLastError());
		AfxMessageBox(_T("Fat32WriteRootDirectory#LocalAlloc Failed"), MB_OK|MB_ICONEXCLAMATION);
		return 1;
	}

	// {[xw肵ƂVolumeIDentryRootDirectory̍ŏɑ}
	if(memcmp(BootSector->VolumeLabel, "NO NAME", 7))
	{
		//set VolumeID entry
		WORD wDate;
		WORD wTime;
		SYSTEMTIME s_time;
		FILETIME f_time;

		GetSystemTime(&s_time);
		//GetLocalTime(&s_time);
		if(!SystemTimeToFileTime(&s_time, &f_time))
		{
			ErrorMessage(GetLastError());
			AfxMessageBox(_T("Fat32WriteRootDirectory#SystemTimeToFileTime Failed"), MB_OK|MB_ICONEXCLAMATION);
		}
		if(!FileTimeToDosDateTime(&f_time, &wDate, &wTime))
		{
			ErrorMessage(GetLastError());
			AfxMessageBox(_T("Fat32WriteRootDirectory#FileTimeToDosDateTime Failed"), MB_OK|MB_ICONEXCLAMATION);
		}

		ZeroMemory(&aDirEntry, sizeof(aDirEntry));
		memcpy_s(aDirEntry.name, 11, BootSector->VolumeLabel, 11);
		aDirEntry.attribute = 0x08; // VolumeID attribute
		*((WORD *)aDirEntry.updateDate) = wDate;
		*((WORD *)aDirEntry.updateTime) = wTime;

		memcpy_s(Buffer, sizeof(aDirEntry), &aDirEntry, sizeof(aDirEntry));
	}

	/* Write cluster */
	FirstDataSector = BootSector->ReservedSectors +
		(BootSector->FATCount * BootSector->FATSectors32) + 0 /* RootDirSectors */;

	FirstRootDirSector = ((BootSector->RootCluster - 2) * BootSector->SectorsPerCluster) + FirstDataSector;
	FileOffset.QuadPart = FirstRootDirSector * BootSector->BytesPerSector;

	dwRet = SetFilePointer(FileHandle ,FileOffset.LowPart, &FileOffset.HighPart, FILE_BEGIN);
	if (dwRet == INVALID_SET_FILE_POINTER && (dwError = GetLastError()) != NO_ERROR ) //changed!!
	{
		ErrorMessage(dwError);
		AfxMessageBox(_T("Fat32WriteRootDirectory#SetFilePointer Failed"), MB_OK|MB_ICONEXCLAMATION);
		return 1;
	}
	bRet = WriteFile(FileHandle, Buffer, BootSector->SectorsPerCluster * BootSector->BytesPerSector, &dwBytes,NULL);
	if(!bRet || dwBytes != BootSector->SectorsPerCluster * BootSector->BytesPerSector)
	{
		ErrorMessage(GetLastError());
		AfxMessageBox(_T("Fat32WriteRootDirectory#WriteFile Failed"), MB_OK|MB_ICONEXCLAMATION);
		LocalFree(Buffer);
		return 1;
	}

	/* Free the buffer */
	LocalFree(Buffer);

	return ERROR_SUCCESS;
}

int Fat32Format::Fat32WriteDataSector(HANDLE FileHandle, PFAT32_BOOT_SECTOR BootSector, CProgressCtrl *pProgFormat)
{
	extern bool g_bStop;
	BOOL bRet;
	DWORD dwRet,dwError,dwBytes =0;
	PUCHAR Buffer;
	ULONG i;
	ULONG Size;
	ULONG Sectors;
	LARGE_INTEGER FileOffset;
	ULONGLONG FirstDataSector;
	ULONGLONG SectorsToFormat;
	DWORD dwWriteUnit = BootSector->SectorsPerCluster * BootSector->BytesPerSector * 512;

	/* Allocate buffer for the cluster */
	Buffer = (PUCHAR)LocalAlloc(LPTR, dwWriteUnit);
	if (Buffer == NULL)
	{
		ErrorMessage(GetLastError());
		AfxMessageBox(_T("Fat32WriteDataSector#LocalAlloc Failed"), MB_OK|MB_ICONEXCLAMATION);
		return 1;
	}

	FirstDataSector = BootSector->ReservedSectors +
		(BootSector->FATCount * BootSector->FATSectors32) + BootSector->SectorsPerCluster;
	SectorsToFormat = BootSector->SectorsHuge - FirstDataSector;
	Sectors = dwWriteUnit / BootSector->BytesPerSector;

	pProgFormat->SetPos(0);
	pProgFormat->SetRange32(0, SectorsToFormat);
	pProgFormat->SetStep(Sectors);

	/* Zero the rest of data sector */
	for (i = FirstDataSector; i < BootSector->SectorsHuge; i += Sectors)
	{
		if(g_bStop)
		{
			LocalFree(Buffer);
			pProgFormat->SetPos(0);
			return 0;
		}

		/* Zero some sectors of the FAT */
		FileOffset.QuadPart = i * BootSector->BytesPerSector;
		Size = BootSector->SectorsHuge - i;
		if (Size > Sectors)
		{
			Size = Sectors;
		}
		Size *= BootSector->BytesPerSector;

		dwRet = SetFilePointer(FileHandle ,FileOffset.LowPart, &FileOffset.HighPart, FILE_BEGIN);
		if (dwRet == INVALID_SET_FILE_POINTER && (dwError = GetLastError()) != NO_ERROR ) //changed!!
		{
			ErrorMessage(dwError);
			AfxMessageBox(_T("Fat32WriteDataSector#SetFilePointer Failed"), MB_OK|MB_ICONEXCLAMATION);
			return 1;
		}
		bRet = WriteFile(FileHandle, Buffer, Size, &dwBytes,NULL);
		if(!bRet || dwBytes != Size)
		{
			ErrorMessage(GetLastError());
			AfxMessageBox(_T("Fat32WriteDataSector#WriteFile Failed"), MB_OK|MB_ICONEXCLAMATION);
			LocalFree(Buffer);
			return 1;
		}

		pProgFormat->StepIt();
	}

	/* Free the buffer */
	LocalFree(Buffer);
	pProgFormat->SetPos(0);

	return ERROR_SUCCESS;
}

int Fat32Format::StartFormat(HANDLE  FileHandle,
			    CString cstrVolumeLabel,
				PPARTITION_INFORMATION  PartitionInfo,
				PDISK_GEOMETRY DiskGeometry,
				BOOL  QuickFormat,
				DWORD  ClusterSize,
				CProgressCtrl *pProgFormat)
{
	FAT32_BOOT_SECTOR BootSector;
	ULONG RootDirSectors;
	ULONG TmpVal1;
	ULONG TmpVal2;
	DWORD dwBytesReturned;
	int nStatus = ERROR_SUCCESS;

	/* Calculate cluster size */
	if (ClusterSize == 0)
	{
		if (PartitionInfo->PartitionLength.QuadPart < 260ULL * 1024ULL * 1024ULL)
		{
			/* Partition < 260MB ==> 0.5KB Cluster */
			ClusterSize = 512;
		}
		if (PartitionInfo->PartitionLength.QuadPart < 8ULL * 1024ULL * 1024ULL * 1024ULL)
		{
			/* Partition < 8GB ==> 4KB Cluster */
			ClusterSize = 4096;
		}
		else if (PartitionInfo->PartitionLength.QuadPart < 16ULL * 1024ULL * 1024ULL * 1024ULL)
		{
			/* Partition 8GB - 16GB ==> 8KB Cluster */
			ClusterSize = 8192;
		}
		else if (PartitionInfo->PartitionLength.QuadPart < 32ULL * 1024ULL * 1024ULL * 1024ULL)
		{
			/* Partition 16GB - 32GB ==> 16KB Cluster */
			ClusterSize = 16384;
		}
		else
		{
			/* Partition >= 32GB ==> 32KB Cluster */
			ClusterSize = 32768;
		}
	}

	memset(&BootSector, 0, sizeof(FAT32_BOOT_SECTOR));
	
	BootSector.magic0 = 0xeb;
	BootSector.res0 = 0x3c;
	BootSector.magic1 = 0x90;
	memcpy(&BootSector.OEMName[0], "MSWIN4.1", 8); //"MSDOS5.0" "MSWIN4.1"
	BootSector.BytesPerSector = DiskGeometry->BytesPerSector;
	BootSector.SectorsPerCluster = ClusterSize / BootSector.BytesPerSector;
	BootSector.ReservedSectors = 32;
	BootSector.FATCount = 2;
	BootSector.RootEntries = 0;
	BootSector.Sectors = 0;
	BootSector.Media = 0xf8;
	BootSector.FATSectors = 0;
	BootSector.SectorsPerTrack = DiskGeometry->SectorsPerTrack;
	BootSector.Heads = DiskGeometry->TracksPerCylinder;
	BootSector.HiddenSectors = PartitionInfo->HiddenSectors;
	BootSector.SectorsHuge = PartitionInfo->PartitionLength.QuadPart >>
		GetShiftCount(BootSector.BytesPerSector); /* Use shifting to avoid 64-bit division */
	BootSector.FATSectors32 = 0; /* Set later */
	BootSector.ExtFlag = 0; /* Mirror all FATs */
	BootSector.FSVersion = 0x0000; /* 0:0 */
	BootSector.RootCluster = 2;
	BootSector.FSInfoSector = 1;
	BootSector.BootBackup = 6;
	BootSector.Drive = 0x08; //0xff/* No BIOS boot drive available */
	BootSector.ExtBootSignature = 0x29;
	BootSector.VolumeID = CalcVolumeSerialNumber ();
	memcpy(&BootSector.VolumeLabel[0], "           ", 11);
	if(cstrVolumeLabel == "")
		memcpy(&BootSector.VolumeLabel[0], "NO NAME    ", 11);
	else
		memcpy(&BootSector.VolumeLabel[0], cstrVolumeLabel.GetBuffer(), cstrVolumeLabel.GetLength());
	memcpy(&BootSector.SysType[0], "FAT32   ", 8);
	BootSector.Signature1 = 0xAA55;

	RootDirSectors = ((BootSector.RootEntries * 32) +
		(BootSector.BytesPerSector - 1)) / BootSector.BytesPerSector;

	/* Calculate number of FAT sectors */
	/* (BytesPerSector / 4) FAT entries (32bit) fit into one sector */
	TmpVal1 = BootSector.SectorsHuge - BootSector.ReservedSectors;
	TmpVal2 = ((BootSector.BytesPerSector / 4) * BootSector.SectorsPerCluster) + BootSector.FATCount;
	BootSector.FATSectors32 = (TmpVal1 + (TmpVal2 - 1)) / TmpVal2;

	if(!DeviceIoControl(
		FileHandle,              // handle to device
		FSCTL_LOCK_VOLUME,  // dwIoControlCode
		NULL,                          // lpInBuffer
		0,                             // nInBufferSize
		NULL,                          // lpOutBuffer
		0,                             // nOutBufferSize
		&dwBytesReturned,     // number of bytes returned
		NULL   // OVERLAPPED structure
		))
	{
		ErrorMessage(GetLastError());
		AfxMessageBox(_T("FSCTL_LOCK_VOLUME failed!"), MB_OK|MB_ICONEXCLAMATION);
		return 1;
	}

	nStatus = Fat32WriteBootSector(FileHandle, &BootSector);
	if (!NT_SUCCESS(nStatus))
	{
		AfxMessageBox(_T("Fat32WriteBootSector() failed!"), MB_OK|MB_ICONEXCLAMATION);
		return 1;
	}

	nStatus = Fat32WriteFsInfo(FileHandle, &BootSector);
	if (!NT_SUCCESS(nStatus))
	{
		AfxMessageBox(_T("Fat32WriteFsInfo() failed!"), MB_OK|MB_ICONEXCLAMATION);
		return 1;
	}

	/* Write first FAT copy */
	nStatus = Fat32WriteFAT(FileHandle, 0, &BootSector);
	if (!NT_SUCCESS(nStatus))
	{
		AfxMessageBox(_T("Fat32WriteFAT()1 failed!"), MB_OK|MB_ICONEXCLAMATION);
		return 1;
	}

	/* Write second FAT copy */
	nStatus = Fat32WriteFAT(FileHandle, BootSector.FATSectors32, &BootSector);
	if (!NT_SUCCESS(nStatus))
	{
		AfxMessageBox(_T("Fat32WriteFAT()2 failed!"), MB_OK|MB_ICONEXCLAMATION);
		return 1;
	}

	nStatus = Fat32WriteRootDirectory(FileHandle, &BootSector);
	if (!NT_SUCCESS(nStatus))
	{
		AfxMessageBox(_T("Fat32WriteRootDirectory() failed!"), MB_OK|MB_ICONEXCLAMATION);
		return 1;
	}

	if (!QuickFormat)
	{
		SetDlgItemText(m_hParentDlg, IDC_STATIC_EXPLANATION, "Initializing data area...");
		nStatus = Fat32WriteDataSector(FileHandle, &BootSector, pProgFormat);
		if (!NT_SUCCESS(nStatus))
		{
			AfxMessageBox(_T("Fat32WriteDataSector() failed!"), MB_OK|MB_ICONEXCLAMATION);
			return 1;
		}
	}

	DeviceIoControl(
		FileHandle,              // handle to device
		FSCTL_UNLOCK_VOLUME,  // dwIoControlCode
		NULL,                          // lpInBuffer
		0,                             // nInBufferSize
		NULL,                          // lpOutBuffer
		0,                             // nOutBufferSize
		&dwBytesReturned,     // number of bytes returned
		NULL   // OVERLAPPED structure
		);

	return 0;
}

void Fat32Format::ErrorMessage(DWORD dwErrorNo)
{
	TCHAR *lpMsgBuf;
	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER|FORMAT_MESSAGE_FROM_SYSTEM,NULL,dwErrorNo, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),(LPTSTR) &lpMsgBuf, 0, NULL);
	AfxMessageBox((LPTSTR)lpMsgBuf, MB_OK|MB_ICONERROR);
	LocalFree(lpMsgBuf);
}